# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Console scripts for OpenTimelineIO

.. moduleauthor:: Contributors to the OpenTimelineIO project <otio-discussion@lists.aswf.io>
"""

# flake8: noqa

# in dependency hierarchy
from . import (
    otioconvert,
    otiocat,
    otiostat,
    otiotool,
    console_utils,
    autogen_serialized_datamodel,
    otiopluginfo,
)

